@@ -22,12 +22,44 @@ static int sort_index;
 static int trim_ws;
 static int show_dictorg_entries;
 
-static void
-memerr(const char *fname)
+static int
+is_alnumspace(unsigned c)
 {
-    dico_log(L_ERR, 0, _("%s: not enough memory"), fname);
+    return utf8_wc_is_alnum(c) || utf8_wc_is_space(c);
 }
 
+static inline int
+headword_compare(char const *a, char const *b, struct dictdb *db)
+{
+    return utf8_compare(a, b,
+			db->flag_casesensitive
+			  ? case_sensitive : case_insensitive,
+			0,
+			db->flag_allchars ? NULL : is_alnumspace);
+}
+
+static inline int
+headword_compare_allchars(char const *a, char const *b, struct dictdb *db,
+			  size_t len)
+{
+    return utf8_compare(a, b,
+			db->flag_casesensitive
+			  ? case_sensitive : case_insensitive,
+			len,
+			NULL);
+}
+
+static int
+compare_index_entry(const void *a, const void *b, void *closure)
+{
+    const struct index_entry *epa = a;
+    const struct index_entry *epb = b;
+    compare_count++;
+    return headword_compare(epa->word, epb->word, (struct dictdb *)closure);
+}
+
+static int get_db_flag(struct dictdb *db, const char *name);
+    
 static int register_strategies(void);
 
 static struct dico_option init_option[] = {
@@ -69,6 +101,13 @@ mod_init(int argc, char **argv)
     return 0;
 }
 
+static void
+dealloc_index_entry(struct index_entry *ent)
+{
+    free(ent->word);
+    free(ent->orig);
+}
+
 static void
 free_db(struct dictdb *db)
 {
@@ -79,7 +118,7 @@ free_db(struct dictdb *db)
     for (i = 0; i < db->numwords; i++) {
 	if (!db->index[i].word)
 	    break;
-	free(db->index[i].word);
+	dealloc_index_entry(&db->index[i]);
     }
     if (db->suf_index) {
 	for (i = 0; i < db->numwords; i++) {
@@ -128,7 +167,6 @@ b64_decode(const char *val, size_t len, size_t *presult)
     return 0;
 }
 
-
 static int
 parse_index_entry(const char *filename, size_t line,
 		  dico_list_t list, char *buf, int tws)
@@ -155,7 +193,8 @@ parse_index_entry(const char *filename, size_t line,
       is used.
     */
     for (nfield = 0; nfield < 4; nfield++) {
-	char *start, *end;
+	char const *start;
+	char const *end;
 	size_t len;
 	
 	/* Skip whitespace */
@@ -181,11 +220,13 @@ parse_index_entry(const char *filename, size_t line,
 		while (len > 0 && start[len-1] == ' ')
 		    --len;
 	    }
+
 	    idx.word = malloc(len + 1);
 	    if (!idx.word) {
-		memerr("parse_index_entry");
+		DICO_LOG_MEMERR();
 		return 1;
 	    }
+
 	    memcpy(idx.word, start, len);
 	    idx.word[len] = 0;
 	    idx.length = len;
@@ -193,7 +234,7 @@ parse_index_entry(const char *filename, size_t line,
 	} else if (nfield == 3) {
 	    idx.orig = malloc(len + 1);
 	    if (!idx.orig) {
-		memerr("parse_index_entry");
+		DICO_LOG_MEMERR();
 		return 1;
 	    }
 	    memcpy(idx.orig, start, len);
@@ -223,7 +264,7 @@ parse_index_entry(const char *filename, size_t line,
 	} else {
 	    ep = malloc(sizeof(*ep));
 	    if (!ep) {
-		memerr("parse_index_entry");
+		DICO_LOG_MEMERR();
 		rc = 1;
 	    }
 	}
@@ -242,8 +283,7 @@ static int
 free_index_entry(void *item, void *data)
 {
     struct index_entry *ep = item;
-    free(ep->word);
-    free(ep->orig);
+    dealloc_index_entry(ep);
     free(ep);
     return 0;
 }
@@ -283,7 +323,7 @@ read_index(struct dictdb *db, const char *idxname, int tws)
 
     list = dico_list_create();
     if (!list) {
-	memerr("read_index");
+	DICO_LOG_MEMERR();
 	rc = 1;
     } else {
 	dico_iterator_t itr;
@@ -333,7 +373,7 @@ open_index(struct dictdb *db, int tws)
     int rc;
     
     if (!idxname) {
-	memerr("open_index");
+	DICO_LOG_MEMERR();
 	return 1;
     }
 
@@ -390,8 +430,6 @@ open_stream(struct dictdb *db)
     return 1;
 }
 
-static int compare_entry(const void *a, const void *b);
-
 static dico_handle_t
 mod_init_db(const char *dbname, int argc, char **argv)
 {
@@ -433,38 +471,52 @@ mod_init_db(const char *dbname, int argc, char **argv)
 	filename = strdup(filename);
 
     if (!filename) {
-	memerr("mod_init_db");
+	DICO_LOG_MEMERR();
 	return NULL;
     }
     
-    db = malloc(sizeof(*db));
+    db = calloc(1, sizeof(*db));
     if (!db) {
 	free(filename);
-	memerr("mod_init_db");
+	DICO_LOG_MEMERR();
 	return NULL;
     }
-    memset(db, 0, sizeof(*db));
 
     db->dbname = dbname;
     db->basename = filename;
     db->show_dictorg_entries = show_dictorg_option;
+    db->flag_allchars = 1;
     
     if (open_index(db, trimws_option)) {
 	free_db(db);
 	return NULL;
     }
 
-    if (sort_option)
-	qsort(db->index, db->numwords, sizeof(db->index[0]), compare_entry);
-    
     if (open_stream(db)) {
 	free_db(db);
 	return NULL;
     }
     
+    db->flag_allchars = get_db_flag(db, DICTORG_FLAG_ALLCHARS);
+    db->flag_casesensitive = get_db_flag(db, DICTORG_FLAG_CASESENSITIVE);
+    db->flag_utf8 = get_db_flag(db, DICTORG_FLAG_UTF8);
+    db->flag_8bit = get_db_flag(db, DICTORG_FLAG_8BIT_NEW);
+    if (get_db_flag(db, DICTORG_FLAG_8BIT_OLD)) {
+	dico_log(L_ERR, 0,
+		 _("mod_init_db(%s): index files in old 8-bit format are not supported"),
+		 argv[0]);
+	free_db(db);
+	return NULL;
+    }
+
+    if (sort_option) {
+	/* Sort index entries */
+	dico_sort(db->index, db->numwords, sizeof(db->index[0]),
+		  compare_index_entry, db);
+    }
+    
     return (dico_handle_t)db;
 }
-
 
 static void
 revert_word(char *dst, const char *src, size_t len)
@@ -484,11 +536,11 @@ revert_word(char *dst, const char *src, size_t len)
 }
 
 static int
-compare_rev_entry(const void *a, const void *b)
+compare_rev_entry(const void *a, const void *b, void *closure)
 {
     struct rev_entry const *epa = a;
     struct rev_entry const *epb = b;
-    return utf8_strcasecmp(epa->word, epb->word);
+    return headword_compare_allchars(epa->word, epb->word, closure, 0);
 }
 
 static int
@@ -512,8 +564,8 @@ init_suffix_index(struct dictdb *db)
 	    db->suf_index[i].word = p;
 	    db->suf_index[i].ptr = &db->index[i];
 	}
-        qsort(db->suf_index, db->numwords, sizeof(db->suf_index[0]),
-	      compare_rev_entry);
+        dico_sort(db->suf_index, db->numwords, sizeof(db->suf_index[0]),
+		  compare_rev_entry, db);
     }
     return 0;
 }    
@@ -523,11 +575,14 @@ static int exact_match(struct dictdb *, const char *, struct result *);
 static int prefix_match(struct dictdb *, const char *, struct result *);
 static int suffix_match(struct dictdb *, const char *, struct result *);
 
-#define RESERVED_WORD(db, word)			            \
-    (!(db)->show_dictorg_entries                            \
-	&& strlen(word) >= sizeof(DICTORG_ENTRY_PREFIX)-1   \
-	&& memcmp(word, DICTORG_ENTRY_PREFIX,               \
-		  sizeof(DICTORG_ENTRY_PREFIX)-1) == 0)
+#define RESERVED_WORD(db, word)					    \
+    (!(db)->show_dictorg_entries				    \
+     && ((strlen(word) >= DICTORG_ENTRY_PREFIX_LEN		    \
+	 && memcmp(word, DICTORG_ENTRY_PREFIX,			    \
+		  DICTORG_ENTRY_PREFIX_LEN) == 0)		    \
+	|| (strlen(word) >= DICTORG_ALT_ENTRY_PREFIX_LEN	    \
+	&& memcmp(word, DICTORG_ALT_ENTRY_PREFIX,		    \
+		      DICTORG_ALT_ENTRY_PREFIX_LEN) == 0)))	    \
 
 
 static struct strategy_def strat_tab[] = {
@@ -556,25 +611,28 @@ register_strategies(void)
 }
 
 static int
-compare_entry(const void *a, const void *b)
-{
-    const struct index_entry *epa = a;
-    const struct index_entry *epb = b;
-    compare_count++;
-    return utf8_strcasecmp(epa->word, epb->word);
-}
-
-static int
-compare_entry_ptr(const void *a, const void *b)
+compare_entry_ptr(const void *a, const void *b, void *closure)
 {
     const struct index_entry *epa = *(const struct index_entry **)a;
     const struct index_entry *epb = *(const struct index_entry **)b;
-    return compare_entry(epa, epb);
+    struct dictdb *db = closure;
+    return headword_compare_allchars(epa->word, epb->word, db, 0);
+}
+
+static int
+uniq_comp(const void *a, const void *b, void *closure)
+{
+    const struct index_entry *epa = a;
+    const struct index_entry *epb = b;
+    struct dictdb *db = closure;
+    /* Prefer original headword over the indexed one. */
+    return headword_compare(epa->orig ? epa->orig : epa->word,
+			    epb->orig ? epb->orig : epb->word, db);
 }
 
 static int
 common_match(struct dictdb *db, const char *word,
-	     int (*compare) (const void *, const void *),
+	     int (*compare)(const void *, const void *, void *),
 	     int unique, struct result *res)
 {
     struct index_entry x, *ep;
@@ -583,32 +641,25 @@ common_match(struct dictdb *db, const char *word,
     x.length = strlen(word);
     x.wordlen = utf8_strlen(word);
     compare_count = 0;
-    ep = bsearch(&x, db->index, db->numwords, sizeof(db->index[0]),
-		 compare);
+    ep = dico_bsearch(&x, db->index, db->numwords, sizeof(db->index[0]),
+		      compare, db);
     if (ep) {
-	struct index_entry *p;
-	for (p = ep - 1; p > db->index && compare(&x, p) == 0; p--)
-	    ;
-	for (ep++; ep < db->index + db->numwords
-		 && compare(&x, ep) == 0; ep++)
-	    ;
 	res->type = result_match;
 	res->db = db;
 	res->list = dico_list_create();
 	if (!res->list) {
-	    memerr("common_match");
+	    DICO_LOG_MEMERR();
 	    return 0;
 	}
 	res->itr = NULL;
 	if (unique) {
-	    dico_list_set_comparator(res->list,
-				     (int (*)(const void *, void *))
-				          compare_entry);
+	    dico_list_set_comparator(res->list, uniq_comp, db);
 	    dico_list_set_flags(res->list, DICO_LIST_COMPARE_TAIL);
 	}
-	for (p++; p < ep; p++) 
-	    if (!RESERVED_WORD(db, p->word))
-		dico_list_append(res->list, p);
+	for (; ep < db->index + db->numwords
+		 && compare(&x, ep, db) == 0; ep++)
+	    if (!RESERVED_WORD(db, ep->word))
+		dico_list_append(res->list, ep);
 	res->compare_count = compare_count;
 	return 0;
     }
@@ -619,19 +670,20 @@ common_match(struct dictdb *db, const char *word,
 static int
 exact_match(struct dictdb *db, const char *word, struct result *res)
 {
-    return common_match(db, word, compare_entry, 1, res);
+    return common_match(db, word, compare_index_entry, 0, res);
 }
 
 static int
-compare_prefix(const void *a, const void *b)
+compare_prefix(const void *a, const void *b, void *closure)
 {
     const struct index_entry *pkey = a;
     const struct index_entry *pelt = b;
+    struct dictdb *db = closure;
     size_t wordlen = pkey->wordlen;
     compare_count++;
     if (pelt->wordlen < wordlen)
 	return -1;
-    return utf8_strncasecmp(pkey->word, pelt->word, wordlen);
+    return headword_compare_allchars(pkey->word, pelt->word, db, wordlen);
 }
 
 static int
@@ -641,15 +693,16 @@ prefix_match(struct dictdb *db, const char *word, struct result *res)
 }
 
 static int
-compare_rev_prefix(const void *a, const void *b)
+compare_rev_prefix(const void *a, const void *b, void *closure)
 {
     const struct rev_entry *pkey = a;
     const struct rev_entry *pelt = b;
+    struct dictdb *db = closure;
     size_t wordlen = pkey->ptr->wordlen;
     if (pelt->ptr->wordlen < wordlen)
 	wordlen = pelt->ptr->wordlen;
     compare_count++;
-    return utf8_strncasecmp(pkey->word, pelt->word, wordlen);
+    return headword_compare_allchars(pkey->word, pelt->word, db, wordlen);
 }
 
 static int
@@ -660,14 +713,14 @@ suffix_match(struct dictdb *db, const char *word, struct result *res)
     int rc;
     
     if (init_suffix_index(db)) {
-	memerr("suffix_match");
+	DICO_LOG_MEMERR();
 	return 1;
     }
     
     ent.length = strlen(word);
     x.word = malloc(ent.length + 1);
     if (!x.word) {
-	memerr("suffix_match");
+	DICO_LOG_MEMERR();
 	return 1;
     }
     ent.wordlen = utf8_strlen(word);
@@ -676,49 +729,42 @@ suffix_match(struct dictdb *db, const char *word, struct result *res)
     x.ptr = &ent;
     
     compare_count = 0;
-    ep = bsearch(&x, db->suf_index, db->numwords, sizeof(db->suf_index[0]),
-		 compare_rev_prefix);
+    ep = dico_bsearch(&x, db->suf_index, db->numwords, sizeof(db->suf_index[0]),
+		      compare_rev_prefix, db);
     if (ep) {
 	struct rev_entry *p;
 	struct index_entry **tmp;
-	size_t i,j;
-	size_t count = 1;
+	size_t i;
+	size_t count = 0;
 	dico_list_t list;
 
-	for (p = ep - 1;
-	     p > db->suf_index && compare_rev_prefix(&x, p) == 0;
-	     p--) 
-	    count++;
-
-	for (ep++;
-	     ep < db->suf_index + db->numwords
-		 && compare_rev_prefix(&x, ep) == 0; ep++)
+	for (p = ep;
+	     p < db->suf_index + db->numwords
+		 && compare_rev_prefix(&x, p, db) == 0; p++)
 	    count++;
 
 	tmp = calloc(count, sizeof(*tmp));
 	if (!tmp) {
-	    memerr("suffix_match");
+	    DICO_LOG_MEMERR();
 	    free(x.word);
 	    return 1;
 	} 
 
-	for (i = j = 0, p++; i < count; i++, p++) 
-	    if (!RESERVED_WORD(db, p->ptr->word)) 
-		tmp[j++] = p->ptr;
+	for (i = 0; i < count; i++) 
+	    if (!RESERVED_WORD(db, ep[i].ptr->word)) 
+		tmp[i] = ep[i].ptr;
 	
-	count = j;
-	qsort(tmp, count, sizeof(tmp[0]), compare_entry_ptr);
+	count = i;
+	dico_sort(tmp, count, sizeof(tmp[0]), compare_entry_ptr, db);
 
 	list = dico_list_create();
 	if (!list) {
-	    memerr("suffix_match");
+	    DICO_LOG_MEMERR();
 	    free(x.word);
 	    free(tmp);
 	    return 1;
 	}
-	dico_list_set_comparator(list,
-				 (int (*)(const void *, void *))
-				   compare_entry);
+	dico_list_set_comparator(list, uniq_comp, db);
 	dico_list_set_flags(list, DICO_LIST_COMPARE_TAIL);
 	for (i = 0; i < count; i++) 
 	    dico_list_append(list, tmp[i]);
@@ -747,13 +793,13 @@ find_db_entry(struct dictdb *db, const char *name)
     x.word = (char*) name;
     x.length = strlen(name);
     x.wordlen = utf8_strlen(name);
-    ep = bsearch(&x, db->index, db->numwords, sizeof(db->index[0]),
-		 compare_entry);
+    ep = dico_bsearch(&x, db->index, db->numwords, sizeof(db->index[0]),
+		      compare_index_entry, db);
     if (!ep)
 	return NULL;
     buf = malloc(ep->size + 1);
     if (!buf) {
-	memerr("find_db_entry");
+	DICO_LOG_MEMERR();
 	return NULL;
     }
     dico_stream_seek(db->stream, ep->offset, DICO_SEEK_SET);
@@ -768,6 +814,18 @@ find_db_entry(struct dictdb *db, const char *name)
     return buf;
 }
 
+static int
+get_db_flag(struct dictdb *db, const char *name)
+{
+    struct index_entry x;
+    
+    x.word = (char*) name;
+    x.length = strlen(name);
+    x.wordlen = utf8_strlen(name);
+    return dico_bsearch(&x, db->index, db->numwords, sizeof(db->index[0]),
+			compare_index_entry, db) != NULL;
+}
+
 static char *
 mod_info(dico_handle_t hp)
 {
@@ -833,13 +891,11 @@ _match_all(struct dictdb *db, dico_strategy_t strat, const char *word)
     list = dico_list_create();
 
     if (!list) {
-	memerr("_match_all");
+	DICO_LOG_MEMERR();
 	return NULL;
     }
 
-    dico_list_set_comparator(list,
-			     (int (*)(const void *, void *))
-			       compare_entry);
+    dico_list_set_comparator(list, uniq_comp, db);
     dico_list_set_flags(list, DICO_LIST_COMPARE_TAIL);
 
     if (dico_key_init(&key, strat, word)) {
@@ -899,12 +955,13 @@ mod_define(dico_handle_t hp, const char *word)
 
     if (RESERVED_WORD(db, word))
 	return NULL;
-    rc = common_match(db, word, compare_entry, 0, &res);
+    
+    rc = common_match(db, word, compare_index_entry, 0, &res);
     if (rc)
 	return NULL;
     rp = malloc(sizeof(*rp));
     if (!rp) {
-	memerr("mod_define");
+	DICO_LOG_MEMERR();
 	dico_list_destroy(&res.list);
 	return NULL;
     }
